from typing import Callable, Union, Any
from collections.abc import Mapping
from nattrs.setter import nested_setattr
from nattrs.getter import nested_getattr
from nattrs.utils import Ignore


def nested_mutattr(
    obj: Union[object, Mapping],
    attr: str,
    fn: Callable,
    is_inplace_fn: bool = False,
    getter_default: Any = Ignore(),
    regex: bool = False,
) -> None:
    r"""
    Mutate object attribute / dict member by recursive lookup, given by dot-separated names.

    Pass `attr='x.a.o'` to mutate attribute "o" of attribute "a" of attribute "x".

    When the attribute doesn't exist, we either ignore the attribute
    (when `getter_default = Ignore()` (default)) or pass the user-specified
    `getter_default` value to the mutating `fn` and create the missing
    attribute.

    Tip: The mutating function can perform checks of typing (e.g. consistency between
    the new and original values) or similar.

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to mutate an attribute/member of a sub-attribute/member of.
        These work interchangeably, why "class, dict, class, ..." work as well.
    attr : str
        The string specifying the dot-separated names of attributes/members.
        The most left name is an attribute/dict member of `obj`
        which has the attribute/key given by the second most left name,
        which has the attribute/key given by the third most left name,
        and so on. The last name may be non-existent and is the name
        of the attribute/member to set.
    fn : Callable
        Function that gets existing attribute/key value and
        returns the new value to be assigned.
    is_inplace_fn : bool
        Whether `fn` affects the attribute inplace. In this case,
        no output is expected (and is ignored), as we expect the
        function to have made the relevant change when applied.
        E.g. useful for `numpy.ndarrays` that we do not
        wish to copy.
        Note: Since no value is directly assigned to `obj`,
        this cannot be done on `getter_default` values
        for missing attributes.
    getter_default : Any
        The value to use for attributes that weren't found.
        Either:
            `Ignore()` (default)
                Leads to no action being taken on those missing attributes.
            User-specified value (e.g. `None`)
                This value is passed to `fn()` and the attribute is created
                with the returned value (note: `is_inplace_fn=True` won't work
                in this case as the value would never be assigned to `obj`).
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. When multiple matches exist, they all
        get mutated.

        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name. Dots within "{}" are respected (i.e. not considered path splits).

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.

    Examples
    --------

    Create class 'B' with a dict 'c' with the member 'd':

    >>> class B:
    >>>     def __init__(self):
    >>>         self.c = {
    >>>             "d": 1
    >>>         }

    Add to a dict 'a':

    >>> a = {"b": B()}

    Mutate `d` with a lambda:

    >>> nested_mutattr(a, "b.c.d", lambda x: x * 5)

    Check new value of d:

    >>> nested_getattr(a, "b.c.d")
    10

    Using regex to find attribute names / dict keys:

    >>> nested_mutattr(a, "b.{.*}.d", lambda x: x * 5, regex=True)

    Check new value of d:

    >>> nested_getattr(a, "b.c.d")
    50  # (i.e. 10 * 5)
    """
    if regex:
        _regex_nested_mutattrs(
            obj=obj,
            attr=attr,
            fn=fn,
            is_inplace_fn=is_inplace_fn,
            getter_default=getter_default,
        )
    else:
        _nested_mutattr(
            obj=obj,
            attr=attr,
            fn=fn,
            is_inplace_fn=is_inplace_fn,
            getter_default=getter_default,
        )


def _nested_mutattr(
    obj: Union[object, Mapping],
    attr: str,
    fn: Callable,
    is_inplace_fn: bool,
    getter_default: Any,
) -> None:
    old_val = nested_getattr(
        obj=obj,
        attr=attr,
        default=getter_default,
    )
    if isinstance(old_val, Ignore):
        return
    try:
        if is_inplace_fn:
            fn(old_val)
        else:
            new_val = fn(old_val)

    except Exception as e:
        raise RuntimeError(
            f"Failed to apply `fn` to the existing value at `{attr}`: {e}"
        )
    if not is_inplace_fn:
        nested_setattr(obj=obj, attr=attr, value=new_val)


def _regex_nested_mutattrs(
    obj: Union[object, Mapping],
    attr: str,
    fn: Callable,
    is_inplace_fn: bool,
    getter_default: Any,
) -> None:
    # Get existing values
    old_vals = nested_getattr(
        obj=obj,
        attr=attr,
        default=getter_default,
        regex=True,
    )

    old_vals = {
        path: old_val
        for path, old_val in old_vals.items()
        if not isinstance(old_val, Ignore)
    }

    try:
        # If the function works in-place
        # e.g. on dicts or objects
        if is_inplace_fn:
            for old_val in old_vals.values():
                fn(old_val)
        else:
            new_vals = {path: fn(old_val) for path, old_val in old_vals.items()}
    except Exception as e:
        raise RuntimeError(
            f"Failed to apply `fn` to the existing value at `{attr}`: {e}"
        )
    if not is_inplace_fn:
        for path, new_val in new_vals.items():
            nested_setattr(
                obj=obj,
                attr=path,
                value=new_val,
                regex="{" in path,
            )



from dataclasses import dataclass
import re
from typing import Dict, List, Tuple, Any
from collections.abc import Mapping, MutableMapping

DOT_PLACEHOLDER = "[[DOT]]"  # Placeholder for dots inside regex terms


def get_or_make_attr(__o: object, __name: str):
    if not hasattr(__o, __name):
        setattr(__o, __name, {})
    return getattr(__o, __name)


def dict_getter(obj: Mapping, key: Any, default: Any):
    if isinstance(default, str):
        if default == "raise":
            return obj[key]
        if default == "make":
            if key not in obj:
                obj[key] = {}
                return obj[key]
    return obj.get(key, default)


def dict_setter(obj: MutableMapping, key: Any, val: Any):
    obj[key] = val


def dict_has(obj: Mapping, key: Any):
    return key in obj


def replace_dots_in_regex(attr: str) -> str:
    """Replace dots inside regex terms (inside `{}`) with a placeholder."""
    result = []
    inside_regex = False

    for char in attr:
        if char == "{":
            inside_regex = True
        elif char == "}":
            inside_regex = False
        if char == "." and inside_regex:
            result.append(DOT_PLACEHOLDER)
        else:
            result.append(char)

    return "".join(result)


def precompile_regexes(terms: List[str]) -> Tuple[str, Dict[str, re.Pattern]]:
    """
    Extract regex patterns wrapped in `{}` that cover entire attribute terms,
    compile them, and replace them with placeholders like `regex_0`, `regex_1`, etc.

    Parameters
    ----------
    terms : List[str]
        List of the attribute names/keys.

    Returns
    -------
    Tuple[str, List[re.Pattern]]
        - The modified dot-joined attribute string with regex placeholders.
        - List of compiled regexes.
    """
    regexes = {}
    new_terms = []

    for i, term in enumerate(terms):
        # Check if the term is completely wrapped in {}
        if term.startswith("{") and term.endswith("}"):
            # Compile the regex inside the curly braces
            pattern = term[1:-1]
            compiled_regex = re.compile(pattern)
            regexes[f"regex_{i}"] = compiled_regex
            # Replace the regex term with a placeholder like 'regex_0'
            new_terms.append("{" + f"regex_{i}" + "}")
        else:
            new_terms.append(term)

    # Join the terms back into a dot-separated path
    modified_attr = ".".join(new_terms)

    return modified_attr, regexes


def extract_terms(attr: str):
    return [
        term.replace(DOT_PLACEHOLDER, ".")
        for term in replace_dots_in_regex(attr).split(".")
    ]


class Ignore:
    """
    Class for indicating that a value
    should be ignored.

    Example
    -------
    >>> if isinstance(x, Ignore):
    >>>     continue

    or

    >>> l = [1, 2, Ignore(), 4]
    >>> for el in l:
    >>>     if not isinstance(el, Ignore):
    >>>         print(el)

    """

    pass


@dataclass(frozen=True)
class MatchResult:
    __slots__ = ["attr_name", "value"]
    attr_name: str
    value: Any


class MissingAttr:
    pass


def get_class_attributes(obj):
    """Return only user-defined attributes, excluding methods and built-in attributes."""
    if hasattr(obj, "__dict__"):
        return obj.__dict__.keys()
    return []


def flatten_matches(lst):
    for item in lst:
        if isinstance(item, list):
            yield from flatten_matches(item)
        else:
            yield item



from collections.abc import Mapping
from typing import Dict, Union

from nattrs.mutator import nested_mutattr
from nattrs.utils import Ignore


def nested_updattr(
    obj: Union[object, Mapping],
    attr: str,
    update_dict: Dict,
    regex: bool = False,
) -> None:
    r"""
    Update object attribute / dict member by recursive lookup.

    Pass `attr='x.a.o'` to update the attribute "o" of attribute "a" of attribute "x".

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to update the attribute/member of.
    attr : str
        The string specifying the dot-separated names of attributes/members.
        When `regex=False`, the attribute must exist or an error is raised.
    update_dict : dict
        Dictionary containing the updates to apply to the target attribute.
        When updating a class instance (i.e. `.__dict__`), keys should be
        valid attribute names (strings with no whitespace etc.)
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. When multiple matches exist, they all
        get updated. Overall non-existing matches are ignored.

        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name. Dots within "{}" are respected (i.e. not considered path splits).

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.
    """

    if not isinstance(update_dict, Mapping):
        raise TypeError(
            "`update_dict` must be of type `Mapping` (e.g. `dict`) "
            f"but was: {type(update_dict)}"
        )

    def update_fn(existing_value):
        if not isinstance(existing_value, dict):
            if hasattr(existing_value, "__dict__"):
                existing_value = existing_value.__dict__
            else:
                raise TypeError(
                    f"Attribute '{attr}' is not a dict or an object with a __dict__"
                )
        existing_value.update(update_dict)
        return existing_value

    # Use nested_mutattr to update the found dictionary
    try:
        nested_mutattr(
            obj=obj,
            attr=attr,
            fn=update_fn,
            is_inplace_fn=True,
            getter_default=Ignore() if regex else None,
            regex=regex,
        )
    except RuntimeError as e:
        if "is not a dict or an object with a __dict__" in str(e):
            raise TypeError(str(e).split(":")[1])
        if "dictionary update sequence element" in str(e):
            raise ValueError(str(e).split(":")[1])
        raise
    except Exception:
        raise



from typing import Union
from collections.abc import Mapping
from nattrs.getter import nested_getattr
import nattrs.utils as utils


def nested_hasattr(
    obj: Union[object, Mapping],
    attr: str,
    allow_none: bool = False,
    regex: bool = False,
):
    r"""
    Check whether recursive object attributes / dict members exist.

    Pass `attr='x.a.o'` to check attribute "o" of attribute "a" of attribute "x".

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to check attributes/members of.
        These work interchangeably, why "class, dict, class, ..." work as well.
    attr : str
        The string specifying the dot-separated names of attributes/members
        to get. The most left name is the object/dict which has
        the attribute/key given by the second most left name,
        which has the attribute/key given by the third most left name,
        and so on.
    allow_none : bool
        Whether to allow `obj` to be `None` (in the first call - non-recursively).
        When allowed, such a call would return `False`.
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. If one or more matches exist,
        the function returns `True`, else `False`.
        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name. Dots within "{}" are respected (i.e. not considered path splits).

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.

        NOTE: When `regex=True`, it simply calls `nested_getattr(..., regex=True)`
        and returns whether anything was found. If you want the found
        attributes, use that directly.

    Returns
    -------
    bool
        Whether (one of) the final attribute/dict member(s) exist.

    Examples
    --------

    Create class 'B' with a dict 'c' with the member 'd':

    >>> class B:
    >>>     def __init__(self):
    >>>         self.c = {
    >>>             "d": 1
    >>>         }

    Add to a dict 'a':

    >>> a = {"b": B()}

    Check presence of the member 'd':

    >>> nested_hasattr(a, "b.c.d")
    True

    Fail to find member 'o':

    >>> nested_hasattr(a, "b.o.p")
    False

    Using regex to find attribute names / dict keys:

    >>> nested_hasattr(a, "b.{.*}.d", regex=True)
    True

    >>> nested_hasattr(a, "b.{.*}.r", regex=True)
    False
    """
    if not allow_none and obj is None:
        raise TypeError("`obj` was `None`.")
    if regex:
        matches = nested_getattr(
            obj,
            attr,
            default=utils.MissingAttr(),
            regex=True,
        )
        return bool(
            [m for m in matches.values() if not isinstance(m, utils.MissingAttr)]
        )

    return _nested_hasattr(obj=obj, attr=attr)


def _nested_hasattr(obj: Union[object, Mapping], attr: str):
    if obj is None:
        return False
    getter = utils.dict_getter if isinstance(obj, Mapping) else getattr
    has_checker = utils.dict_has if isinstance(obj, Mapping) else hasattr
    try:
        left, right = attr.split(".", 1)
    except:  # noqa: E722
        return has_checker(obj, attr)
    return _nested_hasattr(getter(obj, left, None), right)



from .getter import nested_getattr
from .has_checker import nested_hasattr
from .setter import nested_setattr
from .mutator import nested_mutattr
from .deleter import nested_delattr
from .updater import nested_updattr
from .populate import populate_product
from .utils import Ignore


def get_version():
    import importlib.metadata

    return importlib.metadata.version("nattrs")


__version__ = get_version()



from typing import List, Mapping, Union

from nattrs import utils
from nattrs.getter import nested_getattr


def nested_delattr(
    obj: Union[object, Mapping],
    attr: str,
    regex: bool = False,
    allow_missing: bool = True,
) -> None:
    r"""
    Delete object attribute / dict member by recursive lookup, given by dot-separated names.

    Pass `attr='x.a.o'` to delete attribute "o" of attribute "a" of attribute "x".

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to delete an attribute/member of a sub-attribute/member of.
        These work interchangeably, why "class, dict, class, ..." work as well.
    attr : str
        The string specifying the dot-separated names of attributes/members.
        The most left name is an attribute/dict member of `obj`
        which has the attribute/key given by the second most left name,
        which has the attribute/key given by the third most left name,
        and so on. The last name is the one to be deleted.
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. When multiple matches exist, they all
        get deleted.

        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name.  Dots within "{}" are respected (i.e. not considered path splits).

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.

    allow_missing : bool
        Whether to silently skip deletion if the attribute/key does not exist.
        If `False`, a `KeyError` or `AttributeError` will be raised if not found.
        Note: This also silently returns when `obj` is `None`.

    Examples
    --------

    Create class 'B' with a dict 'c' with the member 'd':

    >>> class B:
    >>>     def __init__(self):
    >>>         self.c = {
    >>>             "d": 1,
    >>>             "e": 2,
    >>>         }

    Add to a dict 'a':

    >>> a = {"b": B()}

    Delete the member 'd':

    >>> nested_delattr(a, "b.c.d")

    Check remaining attributes:

    >>> nested_getattr(a, "b.c") == {"e": 2}
    """
    # If obj itself is None, handle it early
    if obj is None:
        if not allow_missing:
            raise AttributeError("`obj` was `None`. Cannot delete from `None` object")
        return
    if regex:
        # Get existing values for regex matching
        matched_vals = nested_getattr(
            obj=obj,
            attr=attr,
            regex=True,
            default=utils.Ignore(),  # Using the "Ignore" sentinel to avoid creating missing values
        )
        for path in matched_vals.keys():
            _regex_nested_delattrs(obj=obj, attr_terms=utils.extract_terms(path))
    else:
        try:
            _nested_delattr(obj=obj, attr=attr, allow_missing=allow_missing)
        except AttributeError as e:
            raise AttributeError("Attribute was not found: ", str(e))
        except KeyError as e:
            raise KeyError("Key was not found: ", str(e))
        except Exception:
            raise


def _nested_delattr(
    obj: Union[object, Mapping],
    attr: str,
    allow_missing: bool = True,
) -> None:
    getter = utils.dict_getter if isinstance(obj, Mapping) else getattr
    try:
        left, right = attr.split(".", 1)
    except ValueError:
        # Final deletion step
        if isinstance(obj, Mapping):
            if allow_missing and attr not in obj:
                return
            del obj[attr]
        else:
            if allow_missing and not hasattr(obj, attr):
                return
            delattr(obj, attr)
        return

    # Get the next attribute level
    next_obj = getter(obj, left, None)
    if obj is None:
        if not allow_missing:
            raise AttributeError(f"nested_delattr(): Attribute ({left}) not found.")
        return  # Early exit if allow_missing=True and next level doesn't exist

    # Recurse to the next level
    _nested_delattr(
        obj=next_obj,
        attr=right,
        allow_missing=allow_missing,
    )


def _regex_nested_delattrs(
    obj: Union[object, Mapping],
    attr_terms: List[str],
    allow_missing: bool = True,
) -> None:
    # No actual regexes at this point,
    # it's just the one for the regex

    # Get left term and remove the
    left_term = attr_terms[0]

    # If nested_getattr() made the attribute
    # into a regex with {} wrapper and escaping of dots
    # we clean it up for fixed lookup
    if left_term.startswith("{") and left_term.endswith("}"):
        left_term = left_term[1:-1].replace(r"\.", ".")

    getter = utils.dict_getter if isinstance(obj, Mapping) else getattr
    if len(attr_terms) > 1:
        right_terms = attr_terms[1:]
    else:
        # Final deletion step
        if isinstance(obj, Mapping):
            if allow_missing and left_term not in obj:
                return
            del obj[left_term]
        else:
            if allow_missing and not hasattr(obj, left_term):
                return
            delattr(obj, left_term)
        return

    # Get the next attribute level
    next_obj = getter(obj, left_term, None)
    if obj is None:
        if not allow_missing:
            raise AttributeError(
                f"nested_delattr(): Attribute ({left_term}) not found."
            )
        return  # Early exit if allow_missing=True and next level doesn't exist

    # Recurse to the next level
    _regex_nested_delattrs(
        obj=next_obj,
        attr_terms=right_terms,
        allow_missing=allow_missing,
    )



import re
from typing import Callable, Dict, List, Optional, Union, Any
from collections.abc import Mapping
from nattrs.utils import Ignore
import nattrs.utils as utils


def nested_getattr(
    obj: Union[object, Mapping],
    attr: str,
    default: Any = None,
    allow_none: bool = False,
    regex: bool = False,
) -> Optional[Union[Any, Dict[str, Any]]]:
    r"""
    Get object attributes / dict members recursively, given by dot-separated names.

    Pass `attr='x.a.o'` to get attribute "o" of attribute "a" of attribute "x".

    Use regular expression to get all matching attribute paths.

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to get attributes/members of.
        These work interchangeably, why "class, dict, class, ..." work as well.
    attr : str
        The string specifying the dot-separated names of attributes/members to get.
        The most left name is an attribute/dict member of `obj`
        which has the attribute/key given by the second most left name,
        which has the attribute/key given by the third most left name,
        and so on.
    default : Any
        Value to return when one or more of the attributes were not found.
        When `regex=True` and you don't want a value for the non-existent
        attributes, pass `default=nattrs.Ignore()` which will be excluded
        from the output. For `regex=False`, any value is simply returned.
    allow_none : bool
        Whether to allow `obj` to be `None` (in the first call - non-recursively).
        When allowed, such a call would return `None`.
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. When multiple matches exist, all are returned
        in a list.

        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name.  Dots within "{}" are respected (i.e. not considered path splits).

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.

    Returns
    -------
    Any or dict[str, Any] or `None`
        When `obj` is `None` and `allow_none` is `True`: `None`

        When `regex=False`: Any
            One of:
                - The value of the final attribute/dict member.
                - The default value

        When `regex=True`: Dict[str, Any]
            Dict mapping matching attribute paths to one of:
                - The respective value of the final attribute/dict member.
                - The default value (except `nattrs.Ignore()` which are excluded).

    Examples
    --------

    Create class 'B' with a dict 'c' with the member 'd':

    >>> class B:
    >>>     def __init__(self):
    >>>         self.c = {
    >>>             "d": 1
    >>>         }

    Add to a dict 'a':

    >>> a = {"b": B()}

    Get the value of 'd':

    >>> nested_getattr(a, "b.c.d")
    1

    Get default value when not finding an attribute:

    >>> nested_getattr(a, "b.o.p", default="not found")
    "not found"

    Using regex to find attribute names / dict keys:

    >>> nested_getattr(a, "b.{.*}.{.*}", regex=True)
    {"b.c.d": 1}

    """
    if obj is None:
        if not allow_none:
            raise TypeError("`obj` was `None`.")
        return None

    # When regex-use is enabled
    if regex:
        # Extract terms for regex compilation
        terms = utils.extract_terms(attr)
        attr, regexes = utils.precompile_regexes(terms)
        matches = _regex_nested_getattrs(
            objs=[obj],
            attr=attr,
            regexes=regexes,
            default=utils.MissingAttr(),
        )

        def _replace_missing(val):
            if isinstance(val, utils.MissingAttr):
                return default
            return val

        return {
            match.attr_name: _replace_missing(match.value)
            for match in utils.flatten_matches(matches)
            if isinstance(match, utils.MatchResult)
            and not isinstance(_replace_missing(match.value), Ignore)
        }

    # Separate method for non-regex version (faster, simpler)
    return _nested_getattr(obj=obj, attr=attr, default=default)


def _nested_getattr(obj: Union[object, Mapping], attr: str, default: Any = None):
    if obj is None:
        return obj
    getter = utils.dict_getter if isinstance(obj, Mapping) else getattr
    try:
        left, right = attr.split(".", 1)
    except:  # noqa: E722
        return getter(obj, attr, default)
    return _nested_getattr(
        obj=getter(obj, left, default),
        attr=right,
        default=default,
    )


def _regex_nested_getattrs(
    objs: List[Union[object, Mapping]],
    attr: str,
    regexes: Dict[str, re.Pattern],
    default: Any,
):
    return [
        _regex_nested_getattr(
            obj=obj,
            attr=attr,
            regexes=regexes,
            default=default,
        )
        for obj in objs
    ]


def _regex_nested_getattr(
    obj: Union[object, Mapping],
    attr: str,
    regexes: Dict[str, re.Pattern],
    default: Any,
):
    prev_attr = ""
    if isinstance(obj, utils.MatchResult):
        prev_attr = obj.attr_name
        obj = obj.value
    if isinstance(obj, utils.MissingAttr):
        return [utils.MissingAttr()]
    if obj is None:
        return [None]

    getter = utils.dict_getter if isinstance(obj, Mapping) else getattr
    try:
        left, right = attr.split(".", 1)
    except:  # noqa: E722
        return _match_and_get_attr(
            obj=obj,
            attr=attr,
            getter=getter,
            regexes=regexes,
            default=default,
            prev_attr=prev_attr,
        )

    return _regex_nested_getattrs(
        _match_and_get_attr(
            obj=obj,
            attr=left,
            getter=getter,
            regexes=regexes,
            default=default,
            prev_attr=prev_attr,
        ),
        attr=right,
        regexes=regexes,
        default=default,
    )


def _match_and_get_attr(
    obj: Any,
    attr: str,
    getter: Callable,
    regexes: Dict[str, re.Pattern],
    default: Any,
    prev_attr: str = "",
) -> List[Any]:
    matches = []

    if attr.startswith("{") and attr.endswith("}"):
        regex_id = attr[1:-1]  # Extract the regex pattern
        pattern = regexes[regex_id]
        keys = (
            obj.keys() if isinstance(obj, Mapping) else utils.get_class_attributes(obj)
        )

        for key in keys:
            if re.fullmatch(pattern, key):
                # Collect all matches
                matches.append(
                    utils.MatchResult(
                        attr_name=f"{prev_attr}.{_dotted_attr(key)}"
                        if prev_attr
                        else key,
                        value=getter(obj, key, default),
                    )
                )
        return (
            matches if matches else [default]
        )  # Return all matches, or default if no match

    # Fallback to non-regex case
    return [
        utils.MatchResult(
            attr_name=f"{prev_attr}.{_dotted_attr(attr)}" if prev_attr else attr,
            value=getter(obj, attr, default),
        )
    ]


def _dotted_attr(s):
    if "." in s:
        return "{" + s.replace(".", r"\.") + "}"
    return s



from functools import partial
from typing import Callable, Dict, List, Any, Optional, Union
from collections.abc import Mapping, MutableMapping
import nattrs.utils as utils
import re


def nested_setattr(
    obj: Union[object, Mapping],
    attr: str,
    value: Any,
    make_missing: bool = False,
    regex: bool = False,
) -> None:
    r"""
    Set object attribute / dict member by recursive lookup, given by dot-separated names.

    Pass `attr='x.a.o'` to set attribute "o" of attribute "a" of attribute "x".

    When `make_missing` is disabled, it requires all but the last attribute/member to already exist.

    If you want type-checking of existing values before assignment (or similar checks),
    consider using `nested_mutattr()` instead and include such checks in the mutator function.

    Note: When the attributes / keys in `attr` themselves include dots, those need to be
    matched with regex patterns (see `regex`).

    Parameters
    ----------
    obj : object (class instance) or dict-like
        The object/dict to set an attribute/member of a sub-attribute/member of.
        These work interchangeably, why "class, dict, class, ..." work as well.
    attr : str
        The string specifying the dot-separated names of attributes/members.
        The most left name is an attribute/dict member of `obj`
        which has the attribute/key given by the second most left name,
        which has the attribute/key given by the third most left name,
        and so on. The last name can be non-existent and is the name
        of the attribute/member to set.
    value : Any
        Value to set for the final attribute/member.
    make_missing : bool
        Whether to create a dict for non-existent intermediate attributes/keys.
        Otherwise, the function will raise either an `AttributeError`
        or a `KeyError` based on the data structure.
    regex : bool
        Whether to interpret attribute/member names wrapped in `{}`
        as regex patterns. When multiple matches exist, they all
        get the value assigned.

        Each regex matching is performed separately per attribute "level"
        (dot separated attribute name).

        Note: The entire attribute name must be included in the wrapper
        (i.e. the first and last character are "{" and "}"),
        otherwise the name is considered a "fixed" (non-regex)
        name. Dots within "{}" are respected (i.e. not considered path splits).

        Note that `make_missing` won't work on attribute names that are
        specified as regex patterns and will fail instead (what would
        be the key?). It will still work on the non-regex names.

        To include an attribute name in `attr` that itself contains a dot,
        use a regex like `r'{x\.y}'` (in context: `r'a.{x\.y}.c'`)
            Remember to escape the dots in the regex or they will
            be considered a regex symbol.


    Examples
    --------

    Create class 'B' with a dict 'c' with the member 'd':

    >>> class B:
    >>>     def __init__(self):
    >>>         self.c = {
    >>>             "d": 1
    >>>         }

    Add to a dict 'a':

    >>> a = {"b": B()}

    Set the value of 'd':

    >>> nested_setattr(a, "b.c.d", 2)

    Check new value of 'd':

    >>> nested_getattr(a, "b.c.d")
    2

    Using regex to find attribute names / dict keys:

    >>> nested_setattr(a, "b.{.*}.d", 5, regex=True)

    Check new value of 'd':

    >>> nested_getattr(a, "b.c.d")
    5

    Set value if it exists or ignore error otherwise:
    >>> try:
    >>>     nested_setattr(a, "b.c.o", "Never set")
    >>> except ValueError:
    >>>     pass

    """

    if regex:
        terms = [
            a.replace(utils.DOT_PLACEHOLDER, ".")
            for a in utils.replace_dots_in_regex(attr).split(".")
        ]
        attr, regexes = utils.precompile_regexes(terms)
        _regex_nested_setattrs(
            objs=[obj],
            attr=attr,
            value=value,
            make_missing=make_missing,
            regexes=regexes,
        )
    else:
        _nested_setattr(
            obj=obj,
            attr=attr,
            value=value,
            make_missing=make_missing,
        )


def _nested_setattr(
    obj: Union[object, Mapping],
    attr: str,
    value: Any,
    make_missing: bool = False,
):
    getter = (
        partial(utils.dict_getter, default="make" if make_missing else "raise")
        if isinstance(obj, Mapping)
        else (utils.get_or_make_attr if make_missing else getattr)
    )

    try:
        left, right = attr.split(".", 1)
    except:  # noqa: E722
        setter = utils.dict_setter if isinstance(obj, MutableMapping) else setattr
        setter(obj, attr, value)
        return
    nested_setattr(
        obj=getter(obj, left),
        attr=right,
        value=value,
        make_missing=make_missing,
    )


def _regex_nested_setattrs(
    objs: List[Union[object, Mapping]],
    attr: str,
    value: Any,
    make_missing: bool,
    regexes: Dict[str, re.Pattern],
) -> None:
    """
    Recursively set attributes for regex-matched paths.
    """
    for obj in objs:
        _regex_nested_setattr(
            obj=obj,
            attr=attr,
            value=value,
            make_missing=make_missing,
            regexes=regexes,
        )


def _regex_nested_setattr(
    obj: Union[object, Mapping],
    attr: str,
    value: Any,
    make_missing: bool,
    regexes: Dict[str, re.Pattern],
) -> None:
    if isinstance(obj, utils.MatchResult):
        obj = obj.value
    if obj is None:
        return [None]

    getter = (
        partial(utils.dict_getter, default="make" if make_missing else "raise")
        if isinstance(obj, Mapping)
        else (
            utils.get_or_make_attr if make_missing else lambda x, y: getattr(x, y, None)
        )
    )

    try:
        left, right = attr.split(".", 1)
    except ValueError:
        setter = utils.dict_setter if isinstance(obj, MutableMapping) else setattr
        _match_and_set_attr(
            obj=obj,
            attr=attr,
            value=value,
            setter=setter,
            getter=getter,
            regexes=regexes,
            make_missing=make_missing,
            is_leaf=True,
        )
        return

    matches = _match_and_set_attr(
        obj=obj,
        attr=left,
        value=value,
        setter=None,
        getter=getter,
        regexes=regexes,
        make_missing=make_missing,
        is_leaf=False,
    )

    _regex_nested_setattrs(
        objs=matches,
        attr=right,
        value=value,
        make_missing=make_missing,
        regexes=regexes,
    )


def _match_and_set_attr(
    obj: Any,
    attr: str,
    value: Any,
    setter: Optional[Callable],
    getter: Callable,
    regexes: Dict[str, re.Pattern],
    make_missing: bool,
    is_leaf: bool,  # Control whether this is a leaf node
    prev_attr: str = "",
) -> List[Any]:
    matches = []

    if attr.startswith("{") and attr.endswith("}"):
        regex_id = attr[1:-1]  # Extract the regex pattern
        pattern = regexes[regex_id]
        keys = (
            obj.keys() if isinstance(obj, Mapping) else utils.get_class_attributes(obj)
        )

        for key in keys:
            if re.fullmatch(pattern, key):
                if is_leaf:  # Only set if this is the final (leaf) term
                    setter(obj, key, value)
                matches.append(
                    utils.MatchResult(
                        attr_name=f"{prev_attr}.{key}" if prev_attr else key,
                        value=getter(obj, key),
                    )
                )
        if make_missing and not matches:
            raise ValueError(
                f"No match was found for regex: '{str(pattern.pattern)}'. "
                f"`make_missing` not possible for regex patterns."
            )
        return matches if matches else [None]

    # For non-regex terms, and if it's a leaf, set the value
    if is_leaf:
        setter(obj, attr, value)

    return [
        utils.MatchResult(
            attr_name=f"{prev_attr}.{attr}" if prev_attr else attr,
            value=getter(obj, attr),
        )
    ]
